--[[********************************************************************
created:	2009/10/10
filename: 	FishTank.lua
author:		Daniel Wellman

purpose:	Simulates a water tank using a Cellular Automata model for mass

License: This software may be reproduced and distributed free of charge as
as long as credit is given to the author. The author assumes no
responsibility. The software is provided as is.
*********************************************************************]]

FishTank = {}

local grid, spoutMass, drain
local WaterProperties = {
	Flow = 5,
	MaxMass = 150,
	MaxCompress = 15,
}

local options = {
	DesignOptions = {"Blizzard", "TownFlood", "FillAndPopTheTop", "Messy", "Empty",},
	Design = "TownFlood",
	SpoutAmount = "Medium",
	SpoutAmountOptions = {"Low", "Medium", "High", "Insane",},
	Drain = "None",
	DrainOptions = {"Bottom", "None",},
}

do--Initialize
	function FishTank:Initialize(width, height)
		local properties = {
			"Inert", 
			"Mass",
			"NewMass",
			"LastTurn",
		}
		grid = Grid:New(width, height, properties)

		for x, y in grid:Iterate() do
			grid:Set(x, y, "Inert", false)
			grid:Set(x, y, "Mass", 0)
			grid:Set(x, y, "NewMass", 0)
			grid:Set(x, y, "LastTurn", false)
		end

		for x, y in Designs:Iterate(options.Design) do
			grid:Set(x, y, "Inert", true)
		end

		if options.SpoutAmount == "Insane" then
			spoutMass = 15000
		elseif options.SpoutAmount == "High" then
			spoutMass = 1500
		elseif options.SpoutAmount == "Medium" then
			spoutMass = 850
		elseif options.SpoutAmount == "Low" then
			spoutMass = 150
		end
	end
	--Add to the menu system
	Menu:RegisterSimulator("FishTank", FishTank, options)

	--Returns the description and mouse controls
	function FishTank:GetDetails()
		return "Simulates a tank of water using cellular automata.", 
			"Spout some water", 
			"Builds or Removes walls"
	end
end



do--Update
	local currentTurn = true
	--We can avoid double buffering by storing the new mass
	--and updating it to the new value as we iterate through
	local function UpdateTurn(x, y)
		if grid(x, y, "LastTurn") ~= currentTurn then
			grid:Set(x, y, "LastTurn", currentTurn)
			grid:Set(x, y, "Mass", grid(x, y, "NewMass"))
		end
	end
	--Clamps a value between two bounds
	local function Clamp(value, max, min)
		if value > max then
			return max
		end
		if value < min then
			return min
		end
		return value
	end
	local HandleInput
	do
		--Input has two distinct modes: placing and removing
		--inertMode saves the state of the mode
		local inertMode
		function HandleInput(dt)
			--Add some water
			if IsLeftMousePressed() then
				local x, y = GetCursorPosition()
				grid:Set(x, y, "NewMass", grid(x, y, "NewMass") + spoutMass * dt * WaterProperties.Flow)
			end
			--Place or remove walls
			if IsRightMousePressed() then
				local x, y = GetCursorPosition()
				if inertMode == nil then
					inertMode = not grid(x, y, "Inert")
				end
				grid:Set(x, y, "Inert", inertMode)
			else
				inertMode = nil
			end
		end
	end
	--When drains are enabled then all cells on the floor lose all their water
	local function HandleDrain(x, y)
		if y == HEIGHT and options.Drain == "Bottom" then
			grid:Set(x, y, "NewMass", 0)
		end
	end
	function FishTank:Update(dt)
		HandleInput(dt)
		--Usually keep locals as close to their definition as possible
		--However, with many cells this micro-optimization results in noticeable speedup
		local flow
		local mass, neighMass
		for x, y in grid:Iterate() do
			if not grid(x, y, "Inert") then
				UpdateTurn(x, y)
				HandleDrain(x, y)
				mass = grid(x, y, "Mass")
				for nx, ny in grid:IterateNeighbors(x, y) do
					if not grid(nx, ny, "Inert") then
						UpdateTurn(nx, ny)
						neighMass = grid(nx, ny, "Mass")
						if ny < y then --Is above?
							if mass < WaterProperties.MaxMass or neighMass < WaterProperties.MaxMass then
								flow = mass - WaterProperties.MaxMass
							else
								flow = mass - neighMass - WaterProperties.MaxCompress
							end
						elseif ny > y then --Is below?
							flow = mass - neighMass + WaterProperties.MaxCompress
						else --Same level
							flow = mass - neighMass
						end
						flow = Clamp(flow * WaterProperties.Flow * dt, mass / 4, -neighMass / 4)

						grid:Set(x, y, "NewMass", grid(x, y, "NewMass") - flow)
						grid:Set(nx, ny, "NewMass", grid(nx, ny, "NewMass") + flow)
					end

				end
			end
		end
		currentTurn = not currentTurn
	end
end

do--Drawing
	local function CellToColor(x, y)
		if grid(x, y, "Inert") then
			return ' ', Colors.Black, Colors.Red
		end
		local pressure = grid(x, y, "Mass")
		if pressure > 2000 then
			return '!', Colors.Yellow, Colors.DarkBlue
		elseif pressure > 100 then
			return ' ', Colors.DarkBlue, Colors.DarkBlue
		elseif pressure > 45 then
			return '*', Colors.Blue, Colors.DarkBlue
		elseif pressure > 35 then
			return ' ', Colors.Blue, Colors.Blue
		elseif pressure > 25 then
			return '*', Colors.DarkTeal, Colors.Blue
		elseif pressure > 15 then
			return ' ', Colors.DarkTeal, Colors.DarkTeal
		elseif pressure > 8 then
			return '*', Colors.Teal, Colors.DarkTeal
		elseif pressure > 4 then
			return ' ', Colors.Teal, Colors.Teal
		elseif pressure > 1 then
			return '*', Colors.Teal, Colors.Black
		elseif pressure > 0.1 then
			return '.', Colors.Teal, Colors.Black
		end
		return ' ', Colors.White, Colors.Black
	end
	function FishTank:Draw()
		for x, y in grid:Iterate() do
			DrawSquare(x, y, CellToColor(x, y))
		end
	end
end