--[[********************************************************************
created:	2009/10/10
filename: 	Grid.lua
author:		Daniel Wellman

purpose:	Implements a NxM matrix using parallel single-dimensional arrays

License: This software may be reproduced and distributed free of charge as
as long as credit is given to the author. The author assumes no
responsibility. The software is provided as is.
*********************************************************************]]

Grid = {}
do
	--Creates a new grid
	function Grid:New(width, height, properties)
		local g = {__Width = width, __Height = height}
		--Create parallel arrays for each property
		for i, prop in ipairs(properties) do
			g[prop] = {}
		end

		--Allow grid to be called to get a property
		setmetatable(g, {__call = Grid.Get,})
		
		--Inject the methods into the grid object
		for name, func in pairs(Grid) do
		    g[name] = func
		end

		return g
	end

	do--IterateNeighbors
		local InitializeIterator
		do
			local Iter
			local currentGrid
			local includeDiagonals
			local function Check(x, y)
				if x <= currentGrid.__Width and x > 0 and y <= currentGrid.__Height and y > 0 then
					return x, y
				end
				return Iter()
				
			end
			local index
			local x, y
			--Elseif chain is fast
			function Iter()
				index = index + 1
				if index == 1 then
					return Check(x-1, y)
				elseif index == 2 then
					return Check(x+1, y)
				elseif index == 3 then
					return Check(x, y+1)
				elseif index == 4 then
					return Check(x, y-1)
				elseif includeDiagonals then
					if index == 5 then
						return Check(x-1, y-1)
					elseif index == 6 then
						return Check(x+1, y-1)
					elseif index == 7 then
						return Check(x-1, y+1)
					elseif index == 8 then
						return Check(x+1, y+1)
					end
				end
			end
			function InitializeIterator(grid, X, Y, full)
				currentGrid = grid
				index = 0
				includeDiagonals = full
				x = X
				y = Y
				return Iter
			end
		end
		--Returns an iterator that iterates through the neighbors around this cell
		function Grid:IterateNeighbors(x, y, full)
			return InitializeIterator(self, x, y, full)
		end
	end

	do--GetNumNeighbors
		--Makes sure the cell is valid and has the property 'propToCheck' passed from GetNumNeighbors
		local function AddNeighbor(self, x, y, prop)
			return x <= self.__Width and x > 0 and y <= self.__Height and y > 0 and 
				self(x, y, prop) and 1 or 0
		end
		function Grid:GetNumNeighbors(x, y, prop)
			return AddNeighbor(self, x-1, y-1, prop) +
			AddNeighbor(self, x-1, y, prop) + 
			AddNeighbor(self, x-1, y+1, prop) +
			AddNeighbor(self, x, y+1, prop) + 
			AddNeighbor(self, x+1, y-1, prop) +
			AddNeighbor(self, x+1, y, prop) + 
			AddNeighbor(self, x+1, y+1, prop) +
			AddNeighbor(self, x, y-1, prop)
		end
	end
	--Iterates through the grid, returning the x and y position
	function Grid:Iterate()
		local width, height = self.__Width, self.__Height
		local x, y = 0, 1
		return function()
			x = x + 1
			if x > width then
				x = 1
				y = y + 1
				if y > height then return end
			end
			return x, y
		end
	end
	
	do--Gets or sets a property
		function Grid:Set(x, y, prop, value)
			self[prop][x + y * self.__Width] = value
		end
		function Grid:Get(x, y, prop)
			return self[prop][x + y * self.__Width]
		end
	end
end